
class ICStationNode extends OptNodeBaseClass
{
	town_mgr = null;
	station_stats = null;
	best_engine_small = null;
	best_engine_large = null;
	
	constructor(station_statistics, town_mgr)
	{
		this.station_stats = station_statistics;
		this.town_mgr = town_mgr;

		// cache the best engine to save some computation time
		this.best_engine_small = Builder.GetEngine(1000, AIVehicle.VT_AIR, true);
		this.best_engine_large = Builder.GetEngine(1000, AIVehicle.VT_AIR, false);
	}

	function GetArchValue(other_node);
	function GetNodeId();
}

function ICStationNode::GetArchValue(other_node)
{
	// Make two lists, one of the airports (ap) and one of the top left tiles of the two airports
	local ap = [this.station_stats.station_id, other_node.station_stats.station_id];
	local ap_tiles = [Station.GetAirportTile(ap[0]), Station.GetAirportTile(ap[1])];
	
	// Check if any of the two airport has a small airport
	local ap_ai_list = AIList();
	ap_ai_list.AddItem(ap[0], 0);
	ap_ai_list.AddItem(ap[1], 0);
	local has_small_airport = Builder.HasSmallAirport(ap_ai_list);

	// best engine for small/large airport have been pre-cached so now we can just get that value
	local best_engine = has_small_airport? this.best_engine_small: this.best_engine_large;

	// Get the distance between the two airports
	local dist_manhattan = AIMap.DistanceManhattan(ap_tiles[0], ap_tiles[1]);

	// Estimate the travel time between the airports using engine: best_engine
	local travel_time = Engine.GetStationToStationAircraftTravelTime(best_engine, ap[0], ap[1]);

	local income = AICargo.GetCargoIncome(Helper.GetPAXCargo(), dist_manhattan, travel_time);

	// Todo: Add a population + pop equity factor
	// or just use the num pax generated last month. 
	local pax_gen = 10;

	return income;
}

function ICStationNode::GetNodeId()
{
	return station_stats.station_id;
}


//// Air Manager ////////////////////////////////////////////////////

class AirManager
{
	static function OptimizeAirportPairs();
}

function AirManager::OptimizeAirportPairs()
{
	local a = 3;
	local pair_optimizer = PairOptimizer();

	local ic_stations = Station.GetAllInterCityStations();
	for(local station_id = ic_stations.Begin(); ic_stations.HasNext(); station_id = ic_stations.Next())
	{
		local town_mgr = towns.GetTownManager(AIStation.GetNearestTown(station_id));
		town_mgr.RemoveInvalidStations();
		town_mgr.AddStationsFromMap();
		local station_stats = town_mgr.GetStationStatistics(station_id);

		if(station_stats != null)
			pair_optimizer.AddNode(ICStationNode(station_stats, town_mgr));
	}

	local solution = pair_optimizer.MakeTabuSearch();

	local i = 1;
	foreach(pair in solution)
	{
		local station1_node = pair.GetNodeObj(0);
		local station2_node = pair.GetNodeObj(1);

		if(station1_node != null)
			Helper.SetSign(AIStation.GetLocation(station1_node.station_stats.station_id) - 1, "Pair " + i);
		if(station2_node != null)
			Helper.SetSign(AIStation.GetLocation(station2_node.station_stats.station_id) - 1, "Pair " + i);

		i++;
	}
}
